/* This is a transient 3D numerical model for contaminant transport
   that uses finite differences for the spatial dimensions and
   forward Euler's method for the temporal. Upwind differences are
   used for the advective terms. This model includes diffusion and
   dispersion (combined) and also advection.

   This program reads FRAC3D input files (nodes, elements, and properties)
   and output files (velocities) plus initial concentrations and then
   marches through time calculating the contaminant transport and
   producing output files that can be read by Tecplot or can be
   converted to TP2 format.

   All of the files for a particular project must begin with the same
   3- to 5-letter prefix and must all have the necessary 3-letter
   extensions (e.g., filename.ext). This is not optional! Spaces are
   not allowed in the prefix and all of the input files must exist
   before launching the program.

   This code was written by Dudley J. Benton
*/

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

void Abort(int code,char*format,...)
  {
  va_list arg_marker;
  fprintf(stderr,"Fatal Error Code %i\n",code);
  va_start(arg_marker,format);
  vfprintf(stderr,format,arg_marker);
  fprintf(stderr,"\n");
  _fcloseall();
  exit(1);
  }

void*allocate(int line,size_t count,size_t size)
  {
  void*ptr;
  if((ptr=calloc(count,size))==NULL)
    Abort(line,"can't allocate memory %ux%u=%u",count,size,count*size);
  return(ptr);
  }

double fmin(double x,double y)
  {
  if(x<y)
    return(x);
  return(y);
  }

double fmax(double x,double y)
  {
  if(x>y)
    return(x);
  return(y);
  }

char cbuf[1024];
char prefix[6];
int*El,*Ie,*Il,*In,*Is,*Iu,*Iw,*La,*Ln,Ne,Nn,Nw;
double Co,rotation;
double dXm,Dxm,Dxx,dYm,Dym,Dyx,dZm,Dzm,Dzx;
double Um,Ux,Vm,Vx,Wm,Wx,Xm,Xx,Ym,Yx,Zm,Zx;

typedef struct{
  int well;  /* extraction well */
  double C;  /* concentration */
  double Dx; /* X diffusion coefficient */
  double Dy; /* Y diffusion coefficient */
  double Dz; /* Z diffusion coefficient */
  double F;  /* porosity */
  double U;  /* X velocity */
  double V;  /* Y velocity */
  double W;  /* Z velocity */
  double X;  /* location */
  double Y;  /* location */
  double Z;  /* location */
  double co; /* cos(rotation) */
  double si; /* sin(rotation) */
  }NODE;
NODE*Node;

typedef struct{
  double a;  /* rotation angle [radians] */
  double Dx; /* X diffusion coefficient */
  double Dy; /* Y diffusion coefficient */
  double Dz; /* Z diffusion coefficient */
  double F;  /* porosity */
  double U;  /* X velocity */
  double V;  /* Y velocity */
  double W;  /* Z velocity */
  double X;  /* location */
  double Y;  /* location */
  double Z;  /* location */
  }ELEM;
ELEM*Elem;

void ReadNodes()
  {
  char fname[13];
  int n;
  double X,Y,Z;
  FILE*fp;
  strcpy(fname,prefix);
  strcat(fname,".NDE");
  printf("node file: %s\n",fname);
  if((fp=fopen(fname,"rt"))==NULL)
    Abort(__LINE__,"can't open essential input file");
  while(1)
    {
    if(!fgets(cbuf,sizeof(cbuf),fp))
      Abort(__LINE__,"unexpected end of file");
    if(cbuf[0]!='*')
      break;
    }
  if(sscanf(cbuf,"%i",&Nn)!=1)
    Abort(__LINE__,"scan error on number of nodes");
  printf("  nodes: %i\n",Nn);
  if(Nn<128)
    Abort(__LINE__,"expected at least 128 nodes");
  Node=allocate(__LINE__,Nn,sizeof(NODE));
  Xm=Ym=Zm=DBL_MAX;
  Xx=Yx=Zx=-DBL_MAX;
  for(n=0;n<Nn;n++)
    {
    while(1)
      {
      if(!fgets(cbuf,sizeof(cbuf),fp))
        Abort(__LINE__,"unexpected end of file");
      if(cbuf[0]!='*')
        break;
      }
    if(sscanf(cbuf,"%lf %lf %lf",&X,&Y,&Z)!=3)
      Abort(__LINE__,"scan error on X Y Z for node %i",n+1);
    Node[n].X=X;
    Node[n].Y=Y;
    Node[n].Z=Z;
    if(X<Xm)
      Xm=X;
    if(X>Xx)
      Xx=X;
    if(Y<Ym)
      Ym=Y;
    if(Y>Yx)
      Yx=Y;
    if(Z<Zm)
      Zm=Z;
    if(Z>Zx)
      Zx=Z;
    }
  fclose(fp);
  printf("  %lGX%lG\n",Xm,Xx);
  printf("  %lGY%lG\n",Ym,Yx);
  printf("  %lGZ%lG\n",Zm,Zx);
  if(Xx<=Xm)
    Abort(__LINE__,"domain is degenerate in X");
  if(Yx<=Ym)
    Abort(__LINE__,"domain is degenerate in Y");
  if(Zx<=Zm)
    Abort(__LINE__,"domain is degenerate in Z");
  }

void ReadElements()
  {
  char fname[13];
  int e,i,j,k,*l,n[8];
  double dx,dy,r,R;
  FILE*fp;
  strcpy(fname,prefix);
  strcat(fname,".ELM");
  printf("element file: %s\n",fname);
  if((fp=fopen(fname,"rt"))==NULL)
    Abort(__LINE__,"can't open essential input file");
  while(1)
    {
    if(!fgets(cbuf,sizeof(cbuf),fp))
      Abort(__LINE__,"unexpected end of file");
    if(cbuf[0]!='*')
      break;
    }
  if(sscanf(cbuf,"%i",&Ne)!=1)
    Abort(__LINE__,"scan error on number of elements");
  printf("  elements: %i\n",Ne);
  if(Ne<32)
    Abort(__LINE__,"expected at least 32 elements");
  El=allocate(__LINE__,Ne*8,sizeof(int));
  Elem=allocate(__LINE__,Ne,sizeof(ELEM));
  l=allocate(__LINE__,Nn,sizeof(int));
  for(k=e=0;e<Ne;e++)
    {
    while(1)
      {
      if(!fgets(cbuf,sizeof(cbuf),fp))
        Abort(__LINE__,"unexpected end of file");
      if(cbuf[0]!='*')
        break;
      }
    if(sscanf(cbuf,"%i %i %i %i %i %i %i %i",n+0,n+1,n+2,n+3,n+4,n+5,n+6,n+7)!=8)
      Abort(__LINE__,"scan error on n1 n2 n3 n4 n5 n6 n7 n8 for element %i",e+1);
    for(i=0;i<8;i++)
      if(n[i]<1||n[i]>Nn)
        Abort(__LINE__,"no such node %i (1 to %i)",n[i],Nn);
    for(i=0;i<7;i++)
      for(j=i+1;j<8;j++)
        if(n[j]==n[i])
          Abort(__LINE__,"element %i contains duplicate node %i",e+1,n[j]);
    for(i=0;i<8;i++)
      {
      El[k++]=n[i]-1;
      l[n[i]-1]++;
      }
    }
  fclose(fp);
  for(i=0;i<Nn;i++)
    if(l[i]==0)
      Abort(__LINE__,"node %i does not appear in any element",i+1);
  R=hypot(Xx-Xm,Yx-Ym);
  for(e=0;e<Ne;e++)
    {
    i=El[8*e];
    j=El[8*e+1];
    dx=Node[j].X-Node[i].X;
    dy=Node[j].Y-Node[i].Y;
    r=hypot(dx,dy);
    if(r<R/1000.)
      Abort(__LINE__,"element %i is too small",e+1);
    rotation=atan2(dy,dx);
    if(e==0)
      printf("  rotation: %.0lf\n",rotation*180./M_PI);
    Elem[e].a=rotation;
    }
  for(k=0;k<Nn;k++)
    l[k]=0;
  for(j=e=0;e<Ne;e++)
    {
    for(i=0;i<8;i++,j++)
      {
      k=El[j];
      Node[k].co+=Elem[e].a;
      l[k]++;
      }
    }
  for(k=0;k<Nn;k++)
    {
    Node[k].co/=l[k];
    Node[k].si=sin(Node[k].co);
    Node[k].co=cos(Node[k].co);
    }
  free(l);
  }

void ReadProperties()
  {
  char fname[13];
  int e;
  double Dx,Dy,Dz,F,Fm,Fx;
  FILE*fp;
  strcpy(fname,prefix);
  strcat(fname,".PRP");
  printf("property file: %s\n",fname);
  if((fp=fopen(fname,"rt"))==NULL)
    Abort(__LINE__,"can't open essential input file");
  while(1)
    {
    if(!fgets(cbuf,sizeof(cbuf),fp))
      Abort(__LINE__,"unexpected end of file");
    if(cbuf[0]!='*')
      break;
    }
  if(sscanf(cbuf,"%i",&e)!=1)
    Abort(__LINE__,"scan error on number of elements");
  if(e!=Ne)
    Abort(__LINE__,"number of elements doesn't match %i?%i",e,Ne);
  Dxm=Dym=Dzm=Fm=DBL_MAX;
  Dxx=Dyx=Dzx=Fx=-DBL_MAX;
  for(e=0;e<Ne;e++)
    {
    while(1)
      {
      if(!fgets(cbuf,sizeof(cbuf),fp))
        Abort(__LINE__,"unexpected end of file");
      if(cbuf[0]!='*')
        break;
      }
    if(sscanf(cbuf,"%lf %lf %lf %lG",&Dx,&Dy,&Dz,&F)!=4)
      Abort(__LINE__,"scan error on Dx Dy Dz F for element %i",e+1);
    if(F<=0.)
      Abort(__LINE__,"porosity0 for element %i",e+1);
    Dx*=365.25;
    Dy*=365.25;
    Dz*=365.25;
    Elem[e].Dx=Dx;
    Elem[e].Dy=Dy;
    Elem[e].Dz=Dz;
    Elem[e].F=F;
    if(Dx<Dxm)
      Dxm=Dx;
    if(Dx>Dxx)
      Dxx=Dx;
    if(Dy<Dym)
      Dym=Dy;
    if(Dy>Dyx)
      Dyx=Dy;
    if(Dz<Dzm)
      Dzm=Dz;
    if(Dz>Dzx)
      Dzx=Dz;
    if(F<Fm)
      Fm=F;
    if(F>Fx)
      Fx=F;
    }
  fclose(fp);
  printf("  %lGDx%lG\n",Dxm,Dxx);
  printf("  %lGDy%lG\n",Dym,Dyx);
  printf("  %lGDz%lG\n",Dzm,Dzx);
  printf("  %lGF%lG\n",Fm,Fx);
  }

void ReadVelocities(double scale)
  {
  char fname[13];
  int e;
  double U,V,W;
  FILE*fp;
  strcpy(fname,prefix);
  strcat(fname,".VEP");
  printf("velocity file: %s\n",fname);
  if((fp=fopen(fname,"rt"))==NULL)
    Abort(__LINE__,"can't open essential input file");
  while(1)
    {
    if(!fgets(cbuf,sizeof(cbuf),fp))
      Abort(__LINE__,"unexpected end of file");
    if(cbuf[0]!='*')
      break;
    }
  if(sscanf(cbuf,"%i",&e)!=1)
    Abort(__LINE__,"scan error on number of elements");
  if(e!=Ne)
    Abort(__LINE__,"number of elements doesn't match %i?%i",e,Ne);
  for(e=0;e<Ne;e++)
    {
    while(1)
      {
      if(!fgets(cbuf,sizeof(cbuf),fp))
        Abort(__LINE__,"unexpected end of file");
      if(cbuf[0]!='*')
        break;
      }
    if(sscanf(cbuf,"%lf %lf %lf",&U,&V,&W)!=3)
      Abort(__LINE__,"scan error on U V W for element %i",e+1);
    U*=scale/Elem[e].F;
    V*=scale/Elem[e].F;
    W*=scale/Elem[e].F;
    Elem[e].U=U;
    Elem[e].V=V;
    Elem[e].W=W;
    if(e==0)
      {
      Um=Ux=U;
      Vm=Vx=V;
      Wm=Wx=W;
      }
    else
      {
      if(U<Um)
        Um=U;
      if(U>Ux)
        Ux=U;
      if(V<Vm)
        Vm=V;
      if(V>Vx)
        Vx=V;
      if(W<Wm)
        Wm=W;
      if(W>Wx)
        Wx=W;
      }
    }
  fclose(fp);
  printf("  %lGU%lG\n",Um,Ux);
  printf("  %lGV%lG\n",Vm,Vx);
  printf("  %lGW%lG\n",Wm,Wx);
  }

void CreateVelocityFile()
  {
  char fname[13];
  int e,i,l,n;
  double X,Y,Z;
  FILE*fp;
  strcpy(fname,prefix);
  strcat(fname,".VEC");
  printf("velocities: %s\n",fname);
  if((fp=fopen(fname,"wt"))==NULL)
    Abort(__LINE__,"can't create output file");
  fprintf(fp,"VARIABLES=\"X\", \"Y\", \"Z\", \"U\", \"V\", \"W\"\n");
  for(l=e=0;e<Ne;e++)
    {
    for(X=Y=Z=i=0;i<8;i++)      /* write element velocities */
      {                         /* calculate element centroids */
      n=El[l++];
      X+=Node[n].X;
      Y+=Node[n].Y;
      Z+=Node[n].Z;
      }
    X/=8.;
    Y/=8.;
    Z/=8.;
    Elem[e].X=X;
    Elem[e].Y=Y;
    Elem[e].Z=Z;
    fprintf(fp,"%lG %lG %lG %lG %lG %lG\n",X,Y,Z,Elem[e].U,Elem[e].V,Elem[e].W);
    }
  fclose(fp);
  }

void AddOne(int n1,int n2)
  {
  int i,l;
  for(l=n1*26,i=0;i<26;i++,l++)
    {
    if(La[l]==n2)
      return;
    else if(La[l]==-1)
      {
      La[l]=n2;
      Ln[n1]++;
      return;
      }
    }
  }

void AddTwo(int n1,int n2)
  {
  AddOne(n1,n2);
  AddOne(n2,n1);
  }

void NodeNeighbors()
  {
  int e,i,j,k,l,m,n;
  printf("neighbors: ");
  La=allocate(__LINE__,Nn*26,sizeof(int));
  Ln=allocate(__LINE__,Nn,sizeof(int));
  Ie=allocate(__LINE__,Nn,sizeof(int));
  Il=allocate(__LINE__,Nn,sizeof(int));
  In=allocate(__LINE__,Nn,sizeof(int));
  Is=allocate(__LINE__,Nn,sizeof(int));
  Iu=allocate(__LINE__,Nn,sizeof(int));
  Iw=allocate(__LINE__,Nn,sizeof(int));
  for(l=e=0;e<Ne;e++)
    {
    for(i=0;i<8;i++,l++)                /* create node properties */
      {                                 /* from element properties */
      n=El[l];
      Node[n].Dx+=Elem[e].Dx;
      Node[n].Dy+=Elem[e].Dy;
      Node[n].Dz+=Elem[e].Dz;
      Node[n].F +=Elem[e].F;
      Node[n].U +=Elem[e].U;
      Node[n].V +=Elem[e].V;
      Node[n].W +=Elem[e].W;
      La[n]++;
      }
    j=8*e;k=El[j];
    m=El[j+1];Ie[k]=m;Iw[m]=k;
    m=El[j+3];In[k]=m;Is[m]=k;
    m=El[j+4];Il[k]=m;Iu[m]=k;
    }
  for(n=0;n<Nn;n++)
    {
    Node[n].Dx/=La[n];
    Node[n].Dy/=La[n];
    Node[n].Dz/=La[n];
    Node[n].F /=La[n];
    Node[n].U /=La[n];
    Node[n].V /=La[n];
    Node[n].W /=La[n];
    }
  for(l=n=0;n<Nn;n++)
    for(i=0;i<26;i++,l++)
      La[l]=-1;
  for(e=0;e<Ne;e++) /* find neighboring nodes */
    for(i=0;i<7;i++)
      for(j=i+1;j<8;j++)
        AddTwo(El[e*8+i],El[e*8+j]);
  for(m=n=0;n<Nn;n++)
    m+=Ln[n];
  printf("%i\n",m);
  for(m=n=0;n<Nn;n++)
    if(Ln[n]==26)
      m++;
  printf("  internal nodes: %i\n",m);
  printf("  boundary nodes: %i\n",Nn-m);
  }

void ReadConcentrations()
  {
  char fname[13];
  int n;
  double C;
  FILE*fp;
  strcpy(fname,prefix);
  strcat(fname,".ICN");
  printf("concentration file: %s\n",fname);
  if((fp=fopen(fname,"rt"))==NULL)
    Abort(__LINE__,"can't open essential input file");
  while(1)
    {
    if(!fgets(cbuf,sizeof(cbuf),fp))
      Abort(__LINE__,"unexpected end of file");
    if(cbuf[0]!='*')
      break;
    }
  if(sscanf(cbuf,"%i",&n)!=1)
    Abort(__LINE__,"scan error on number of nodes");
  if(n!=Nn)
    Abort(__LINE__,"number of nodes doesn't match %i?%i",n,Nn);
  for(n=0;n<Nn;n++)
    {
    while(1)
      {
      if(!fgets(cbuf,sizeof(cbuf),fp))
        Abort(__LINE__,"unexpected end of file");
      if(cbuf[0]!='*')
        break;
      }
    if(sscanf(cbuf,"%lf",&C)!=1)
      Abort(__LINE__,"scan error on concentration for node %i",n+1);
    if(C<0.)
      Abort(__LINE__,"concentration for node %i is <0",n+1);
    if(C>Co)
      Co=C;
    Node[n].C=C;
    if(C>Co)
      Co=C;
    }
  fclose(fp);
  printf("  C%lG\n",Co);
  if(Co<=0.)
    Abort(__LINE__,"there is nothing to transport");
  }

void AddWell(double X,double Y,double R)
  {
  int i,n;
  Nw++;
  for(i=n=0;n<Nn;n++)
    if(hypot(Node[n].X-X,Node[n].Y-Y)<=R)
      {
      Node[n].well=Nw;
      i++;
      }
  printf("%i nodes in well %i\n",i,Nw);
  }

int relax=0;

double AdvanceSolution(double dt)
  {
  int i,l,n,r;
  double Ca,Cb,Cc,Cd,Ce,Cf,Cg,Ch,Cl,Cn,co,Cs,Cu,Cw,Cx,dP,Dx,dQ,Dy,dZ,Dz,si,U,V,W;
  double d2CdP2,d2CdQ2,d2CdZ2,dCdPa,dCdPb,dCdQd,dCdQf,dCdZg,dCdZh,UdCdX,VdCdY,WdCdZ;
  static double*dCdt;
  if(dCdt==NULL)
    dCdt=allocate(__LINE__,Nn,sizeof(double));
  for(n=0;n<Nn;n++)
    {
    if(Ln[n]<26)                /* only consider internal nodes */
      continue;
    dP=hypot(Node[Ie[n]].Y-Node[Iw[n]].Y,Node[Ie[n]].X-Node[Iw[n]].X)/2.;
    dQ=hypot(Node[In[n]].Y-Node[Is[n]].Y,Node[In[n]].X-Node[Is[n]].X)/2.;
    dZ=(Node[Iu[n]].Z-Node[Il[n]].Z)/2.;
    Cc=Node[n].C;
    Ce=Node[Ie[n]].C;
    Cn=Node[In[n]].C;
    Cs=Node[Is[n]].C;
    Cw=Node[Iw[n]].C;
    Cu=Node[Iu[n]].C;
    Cl=Node[Il[n]].C;
    Ca=(Cw+Cc)/2.;
    Cb=(Ce+Cc)/2.;
    Cd=(Cn+Cc)/2.;
    Cf=(Cs+Cc)/2.;
    Cg=(Cu+Cc)/2.;
    Ch=(Cl+Cc)/2.;
    dCdPa=(Cc-Cw)/dP;
    dCdPb=(Ce-Cc)/dP;
    dCdQd=(Cn-Cc)/dQ;
    dCdQf=(Cc-Cs)/dQ;
    dCdZg=(Cu-Cc)/dZ;
    dCdZh=(Cc-Cl)/dZ;
    d2CdP2=(dCdPb-dCdPa)/dP;
    d2CdQ2=(dCdQd-dCdQf)/dQ;
    d2CdZ2=(dCdZg-dCdZh)/dZ;
    Dx=Node[n].Dx;
    Dy=Node[n].Dy;
    Dz=Node[n].Dz;
    U=Node[n].U;
    V=Node[n].V;
    W=Node[n].W;
    co=Node[n].co;
    si=Node[n].si;
    if(U>0.)
      UdCdX=U*(co*(Cc-Cw)/dP-si*(Cc-Cs)/dQ);
    else
      UdCdX=U*(co*(Ce-Cc)/dP-si*(Cn-Cc)/dQ);
    if(V>0.)
      VdCdY=V*(si*(Cc-Cw)/dP+co*(Cc-Cs)/dQ);
    else
      VdCdY=V*(si*(Ce-Cc)/dP+co*(Cn-Cc)/dQ);
    if(W>0.)
      WdCdZ=W*(Cc-Cl)/dZ;
    else
      WdCdZ=W*(Cu-Cc)/dZ;
    dCdt[n]=Dx*(co*d2CdP2-si*d2CdQ2)
           +Dy*(si*d2CdP2+co*d2CdQ2)
           +Dz*d2CdZ2
           -UdCdX
           -VdCdY
           -WdCdZ;
    }
  for(r=0;r<relax;r++)
    {
    for(n=0;n<Nn;n++)
      {
      for(i=0;i<Ln[n];i++)
        {
        l=La[26*n+i];
        dCdt[n]+=dCdt[l];
        }
      dCdt[n]/=Ln[n]+1;
      }
    }
  for(Cx=n=0;n<Nn;n++)
    {
    Node[n].C+=dt*dCdt[n];
    if(Node[n].C>Cx)
      Cx=Node[n].C;
    }
  return(Cx);
  }

double DetermineTimeStep()
  {
  int e,i,l,n;
  double atx,aty,atz,dt,dtx,dty,dtz,dX,dY,dZ,X,xm,xx,Y,ym,yx,Z,zm,zx;
  printf("determining time step\n");
  dXm=dYm=dZm=DBL_MAX;
  for(l=e=0;e<Ne;e++)
    {
    xm=ym=zm=DBL_MAX;
    xx=yx=zx=-DBL_MAX;
    for(i=0;i<8;i++,l++)
      {
      n=El[l];
      X=Node[n].X;
      Y=Node[n].Y;
      Z=Node[n].Z;
      if(X<xm)
        xm=X;
      if(X>xx)
        xx=X;
      if(Y<ym)
        ym=Y;
      if(Y>yx)
        yx=Y;
      if(Z<zm)
        zm=Z;
      if(Z>zx)
        zx=Z;
      }
    dX=xx-xm;
    dY=yx-ym;
    dZ=zx-zm;
    if(dX<(Xx-Xm)/1000.)
      Abort(__LINE__,"element %i is degenerate in X",e+1);
    if(dY<(Yx-Ym)/1000.)
      Abort(__LINE__,"element %i is degenerate in Y",e+1);
    if(dZ<(Zx-Zm)/1000.)
      Abort(__LINE__,"element %i is degenerate in Z",e+1);
    if(dX<dXm)
      dXm=dX;
    if(dY<dYm)
      dYm=dY;
    if(dZ<dZm)
      dZm=dZ;
    }
  printf("  X=%lG, Y=%lG, Z=%lG\n",dX,dY,dZ);
  dtx=dX*dX/fmin(fabs(Dxm),fabs(Dxx))/2.;
  dty=dY*dY/fmin(fabs(Dym),fabs(Dyx))/2.;
  dtz=dZ*dZ/fmin(fabs(Dzm),fabs(Dzx))/2.;
  printf("  X/Dx/2=%lG, Y/Dy/2=%lG, Z/Dz/2=%lG\n",dtx,dty,dtz);
  atx=dX/fmax(fabs(Um),fabs(Ux))/6.;
  aty=dY/fmax(fabs(Vm),fabs(Vx))/6.;
  atz=dZ/fmax(fabs(Wm),fabs(Wx))/6.;
  printf("  X/U/6=%lG, Y/V/6=%lG, Z/W/6=%lG\n",atx,aty,atz);
  dt=1./(1./dtx+1./dty+1./dtz+1./atx+1./aty+1./atz);
  printf("  t=%lG\n",dt);
  return(dt);
  }

void FileResults(char*name)
  {
  int e,i,l,n;
  double C,Cm,logCm;
  FILE*fp;
  printf("%s\n",name);
  if((fp=fopen(name,"wt"))==NULL)
    Abort(__LINE__,"can't create output file");
  fprintf(fp,"VARIABLES=\"X\", \"Y\", \"Z\", \"C\"\n");
  fprintf(fp,"ZONE N=%i E=%i F=FEPOINT ET=BRICK\n",Nn,Ne);
  Cm=Co/1E6;
  logCm=log(Cm);
  for(n=0;n<Nn;n++)
    {
    C=Node[n].C;
    if(C<Cm)
      C=logCm;
    else
      C=log(C);
    fprintf(fp,"%lG %lG %lG %lG\n",Node[n].X,Node[n].Y,Node[n].Z,C);
    }
  for(l=e=0;e<Ne;e++)
    {
    for(i=0;i<8;i++)
      fprintf(fp," %i",El[l++]+1);
    fprintf(fp,"\n");
    }
  fclose(fp);
  }

void CheckPrefix(int argc,char**argv)
  {
  if(argc<2)
    Abort(__LINE__,"no input specified... try\nctmod wine");
  if(argc>2)
    Abort(__LINE__,"too many parameters specified... try\nctmod wine");
  printf("prefix: %s\n",argv[1]);
  if(strlen(argv[1])<3)
    Abort(__LINE__,"prefix must be at least 3 characters");
  if(strlen(argv[1])>5)
    Abort(__LINE__,"prefix must be no more than 5 characters");
  if(strchr(argv[1],' '))
    Abort(__LINE__,"prefix must not contain blanks");
  if(strchr(argv[1],'.'))
    Abort(__LINE__,"prefix must not contain a dot or have an extension");
  strcpy(prefix,argv[1]);
  if(strspn(prefix,"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_0123456789")!=strlen(prefix))
    Abort(__LINE__,"prefix contains illegal character (must be A-Za-z_0-9)");
  _strupr(prefix);
  }

int main(int argc,char**argv,char**envp)
  {
  char fname[13];
  int i,j,k=0,n;
  double Cx,dt,t;
  printf("3D Transient Diffusion + Dispersion + Advection\n");
  printf("reading FRAC3D input and output files\n");
  CheckPrefix(argc,argv);
  ReadNodes();
  ReadElements();
  ReadProperties();
  ReadVelocities(1.);
  dt=DetermineTimeStep();
  if(dt>1.)
    dt=1.;
  NodeNeighbors();
  AddWell(0.,100.,5.);
restart:
  ReadConcentrations();
  printf("0 years C%.0lf ",Co);
  sprintf(fname,"%s000.DAT",prefix);
  FileResults(fname);
  n=(int)(1./dt+0.5);
  for(t=j=i=0;t<1800.;i++)
    {
    Cx=AdvanceSolution(dt);
    if(Cx>Co*2.)
      {
      printf("\rCx=%lG\ntime step is too large - solution is unstable\n",Cx);
      if(++k>4)
        break;
      dt/=2.;
      relax++;
      printf("restart with t=%lG, relax=%i\n",dt,relax);
      goto restart;
      }
    t+=dt;
    printf("%.3lf years C%.0lf ",t,Cx);
    if((i+1)%n==0)
      {
      sprintf(fname,"%s%03d.DAT",prefix,++j);
      FileResults(fname);
      if(Cx<Co/100.)
        break;
      }
    else
      printf("\r");
    }
  return(0);
  }
