/* This is a transient 3D numerical model for contaminant transport
   that uses finite differences for the spatial dimensions and
   forward Euler's method for the temporal. Upwind differences are
   used for the advective terms. This model includes diffusion and
   dispersion (combined) and also advection.

   This program reads MODFLOW input files (nodes, elements, and properties)
   and output files (velocities) plus initial concentrations and then
   marches through time calculating the contaminant transport and
   producing output files that can be read by TP2 or converted to
   Tecplot format.

   All of the files for a particular project must begin with the same
   3- to 5-letter prefix and must all have the necessary 3-letter
   extensions (e.g., filename.ext). This is not optional! Spaces are
   not allowed in the prefix and all of the input files must exist
   before launching the program.

   This code was written by Dudley J. Benton
*/

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

void Abort(int code,char*format,...)
  {
  va_list arg_marker;
  fprintf(stderr,"Fatal Error Code %i\n",code);
  va_start(arg_marker,format);
  vfprintf(stderr,format,arg_marker);
  fprintf(stderr,"\n");
  _fcloseall();
  exit(1);
  }

void*allocate(int line,size_t count,size_t size)
  {
  void*ptr;
  if((ptr=calloc(count,size))==NULL)
    Abort(line,"can't allocate memory %ux%u=%u",count,size,count*size);
  return(ptr);
  }

char*stristr(char*str1,char*str2)        /* find string in string-ignore case */
  {
  int l1,l2;
  l1=(int)strlen(str1);
  l2=(int)strlen(str2);
  l1-=l2;
  while(l1>=0)
    {
    if(_memicmp(str1,str2,l2))
      str1++;
    else
      return(str1);
    l1--;
    }
  return(NULL);
  }

char*strtrm(char*string)            /* trim string (remove extraneous blanks) */
  {
  int i;
  i=0;
  while(string[i]==' ')
    i++;
  if(i)
    strcpy(string,string+i);
  i=(int)strlen(string);
  while(i>0&&string[i-1]==' ')
    i--;
  string[i]=0;
  return(string);
  }

double fmin(double x,double y)
  {
  if(x<y)
    return(x);
  return(y);
  }

double fmax(double x,double y)
  {
  if(x>y)
    return(x);
  return(y);
  }

char cbuf[1024];
char prefix[6];
int Ncol; /* number of columns of cells */
int Ne;   /* number of elements */
int Nlay; /* number of layers of cells */
int Nn;   /* number of nodes */
int Nrow; /* number of rows of cells */
int*El;   /* list of nodes for each element (8 per brick) */
float*dC; /* colum spacing */
float*dL; /* layer spacing */
float*dR; /* row spacing */
double tau=1E6; /* exponential decay factor [years] */
double Co,*dCdt,Fmin,Fmax,Dxmn,Dxmx,Dymn,Dymx,Dzmn,Dzmx;
double Umin,Umax,Vmin,Vmax,Wmin,Wmax;
double Xmin,Xmax,Xoff,Ymin,Ymax,Yoff,Zmin,Zmax,Ysiz,Xsiz,Zsiz;

typedef struct{
  int active;/* active/inactive flag */
  double C;  /* concentration */
  double X;  /* location */
  double Y;  /* location */
  double Z;  /* location */
  }NODE;
NODE*Node;

typedef struct{
  int active;/* active/inactive flag */
  double Dx; /* X diffusion coefficient */
  double Dy; /* Y diffusion coefficient */
  double Dz; /* Z diffusion coefficient */
  double F;  /* porosity */
  double U;  /* X velocity */
  double V;  /* Y velocity */
  double W;  /* Z velocity */
  double X;  /* X centroid */
  double Y;  /* Y centroid */
  double Z;  /* Z centroid */
  }ELEM;
ELEM*Elem;

int*ReadIntegers(FILE*fp,int n)
  {
  int posi;
  int form,i,l,*Lint;
  Lint=allocate(__LINE__,n,sizeof(int));
  fgets(cbuf,sizeof(cbuf),fp);
  if(sscanf(cbuf,"%i%i",&form,&l)!=2)
    Abort(__LINE__,"scan error reading data input format index");
  if(form>0)
    {
    cbuf[0]=0;
    for(i=0;i<n;)
      {
      if(sscanf(cbuf,"%i%n",Lint+i,&posi)==1)
        {
        i++;
        strcpy(cbuf,cbuf+posi);
        }
      else
        fgets(cbuf,sizeof(cbuf),fp);
      }
    }
  else if(form==0)
    {
    for(i=0;i<n;i++)
      Lint[i]=l;
    }
  else
    Abort(__LINE__,"can't handle binary input for MODFLOW files");
  return(Lint);
  }

float*ReadReals(FILE*fp,int n)
  {
  int posi;
  int form,i;
  float F,*Freal;
  Freal=allocate(__LINE__,n,sizeof(float));
  fgets(cbuf,sizeof(cbuf),fp);
  if(sscanf(cbuf,"%i%f",&form,&F)!=2)
    Abort(__LINE__,"scan error reading data input format index");
  if(form>0)
    {
    cbuf[0]=0;
    for(i=0;i<n;)
      {
      if(sscanf(cbuf,"%f%n",Freal+i,&posi)==1)
        {
        i++;
        strcpy(cbuf,cbuf+posi);
        }
      else
        fgets(cbuf,sizeof(cbuf),fp);
      }
    }
  else if(form==0)
    {
    for(i=0;i<n;i++)
      Freal[i]=F;
    }
  else
    Abort(__LINE__,"can't handle binary input for MODFLOW files");
  return(Freal);
  }

int node(int x,int y,int z)
  {
  return((x*(Nrow+1)+y)*(Nlay+1)+z);
  }

int elem(int x,int y,int z)
  {
  return((x*Nrow+y)*Nlay+z);
  }

void ReadBasicInputOptionFile()
  {
  char fname[13];
  int*act,e,i,n,x,y,z;
  FILE*fp;
  strcpy(fname,prefix);
  strcat(fname,".BAS");
  printf("basic input file: %s\n",fname);
  if((fp=fopen(fname,"rt"))==NULL)
    Abort(__LINE__,"can't open essential input file");
  fgets(cbuf,sizeof(cbuf),fp);
  strtrm(cbuf);
  printf("  title: %s",cbuf);
  fgets(cbuf,sizeof(cbuf),fp);
  strtrm(cbuf);
  printf("  subtitle: %s",cbuf);
  if(stristr(cbuf,"reference:"))               /* scan for optional reference */
    {
    if(sscanf(stristr(cbuf,"reference:")+10,"%lf%*[ ,]%lf",&Xoff,&Yoff)==2)
      printf("\n  Reference Point: X=%lG, Y=%lG\n",Xoff,Yoff);
    else
      Abort(__LINE__,"scan error on reference point in .BAS file");
    }
  else
    Xoff=Yoff=0.;
  fgets(cbuf,sizeof(cbuf),fp);
  if(sscanf(cbuf,"%i%i%i",&Nlay,&Nrow,&Ncol)!=3)
    Abort(__LINE__,"scan error reading: Nlay,Nrow,Ncol");
  printf("  grid: %ix%ix%i\n",Ncol,Nrow,Nlay);
  if(Ncol<2)
    Abort(__LINE__,"Ncol<2");
  if(Nrow<2)
    Abort(__LINE__,"Nrow<2");
  if(Nlay<2)
    Abort(__LINE__,"Nlay<2");
  Ne=Ncol*Nrow*Nlay;
  printf("  total cells: %i\n",Ne);
  Elem=allocate(__LINE__,Ne,sizeof(ELEM));
  fgets(cbuf,sizeof(cbuf),fp);
  fgets(cbuf,sizeof(cbuf),fp);
  for(n=0,z=Nlay-1;z>=0;z--)    /* read top-to-bottom */
    {
    act=ReadIntegers(fp,Ncol*Nrow);
    for(i=0,y=Nrow-1;y>=0;y--)  /* read north-to-south */
      {
      for(x=0;x<Ncol;x++,i++)   /* read west-to-east */
        {
        e=elem(x,y,z);
        if(act[i])
          {
          Elem[e].active=1;
          n++;
          }
        }
      }
    free(act);
    }
  fclose(fp);
  printf("  active cells: %i\n",n);
  printf("  augmented grid: %ix%ix%i\n",Ncol+1,Nrow+1,Nlay+1);
  printf("  element type: hexahedra (bricks)\n");
  Nn=(Ncol+1)*(Nrow+1)*(Nlay+1);
  printf("  resulting nodes: %i\n",Nn);
  Node=allocate(__LINE__,Nn,sizeof(NODE));
  dCdt=allocate(__LINE__,Nn,sizeof(double));
  El=allocate(__LINE__,Ne*8,sizeof(int));
  for(i=e=x=0;x<Ncol;x++)
    {
    for(y=0;y<Nrow;y++)
      {
      for(z=0;z<Nlay;z++,e++)
        {
        El[i++]=node(x  ,y  ,z  );
        El[i++]=node(x+1,y  ,z  );
        El[i++]=node(x+1,y+1,z  );
        El[i++]=node(x  ,y+1,z  );
        El[i++]=node(x  ,y  ,z+1);
        El[i++]=node(x+1,y  ,z+1);
        El[i++]=node(x+1,y+1,z+1);
        El[i++]=node(x  ,y+1,z+1);
        }
      }
    }
  for(n=e=0;e<Ne;e++)
    for(i=0;i<8;i++,n++)
      if(Elem[e].active)
        Node[El[n]].active=1;
  for(i=n=0;n<Nn;n++)
    if(Node[n].active)
      i++;
  printf("  active nodes: %i\n",i);
  }

void ReadBlockCenteredFile()
  {
  char fname[13];
  int c,e,i,k,l,n,r,s,x,y,z;
  float*discard,*Zbot,*Zswp,*Ztop;
  double X,Y,Z;
  FILE*fp;
  strcpy(fname,prefix);
  strcat(fname,".BCF");
  printf("block-centered file: %s\n",fname);
  if((fp=fopen(fname,"rt"))==NULL)
    Abort(__LINE__,"can't open essential input file");
  fgets(cbuf,sizeof(cbuf),fp);
  fgets(cbuf,sizeof(cbuf),fp);
  for(l=k=0;l<Nlay;l++)
    {
    if(sscanf(cbuf+k,"%i%n",&i,&n)!=1)
      Abort(__LINE__,"scan error reading LAYCON");
    if(i!=3&&i!=43)
      Abort(__LINE__,"LAYCON must be 3 or 43 for every layer; otherwise Zelem is unknown");
    k+=n;
    }
  discard=ReadReals(fp,Nlay);
  free(discard);
  dC=ReadReals(fp,Ncol);
  for(c=0,Xmin=Xmax=Xoff;c<Ncol;c++)
    if(dC[c]>0.)
      Xmax+=dC[c];
    else
      Abort(__LINE__,"delta-C0");
  for(c=0,X=Xmin;c<=Ncol;X+=dC[c++])
    {
    x=c;
    for(r=0;r<=Nrow;r++)
      {
      y=Nrow-r;
      for(l=0;l<=Nlay;l++)
        {
        z=Nlay-l;
        Node[node(x,y,z)].X=X;
        }
      }
    }
  printf("  %lGX%lG\n",Xmin,Xmax);
  dR=ReadReals(fp,Nrow);
  for(r=0,Ymin=Ymax=Yoff;r<Nrow;r++)
    if(dR[r]>0.)
      Ymax+=dR[r];
    else
      Abort(__LINE__,"delta-R0");
  for(r=0,Y=Ymax;r<=Nrow;Y-=dR[r++])
    {
    y=Nrow-r;
    for(c=0;c<=Ncol;c++)
      {
      x=c;
      for(l=0;l<=Nlay;l++)
        {
        z=Nlay-l;
        Node[node(x,y,z)].Y=Y;
        }
      }
    }
  printf("  %lGY%lG\n",Ymin,Ymax);
  dL=allocate(__LINE__,Ne,sizeof(float));
  for(l=0;l<Nlay;l++)
    {
    discard=ReadReals(fp,Ncol*Nrow);
    free(discard);
    Zbot=ReadReals(fp,Ncol*Nrow);
    if(l<Nlay-1)                              /* Note: no Kz for bottom layer */
      {
      discard=ReadReals(fp,Ncol*Nrow);
      free(discard);
      }
    Ztop=ReadReals(fp,Ncol*Nrow);
    z=Nlay-l-1;
    for(r=i=0;r<Nrow;r++)
      {
      y=Nrow-r-1;
      for(c=0;c<Ncol;c++,i++)
        {
        x=c;
        if(Ztop[i]<=Zbot[i])
          Abort(__LINE__,"cell depth0 layer=%i, row=%i, col=%i",z+1,y+1,x+1);
        dL[elem(x,y,z)]=Ztop[i]-Zbot[i];
        }
      }
    for(s=0;s<2;s++)
      {
#define ZTOP(r,c) Ztop[(r)*Ncol+c]
      z=Nlay-l-s;
      for(r=0;r<=Nrow;r++)
        {
        y=Nrow-r;
        for(c=0;c<=Ncol;c++)
          {
          x=c;
          if(c==0)
            {
            if(r==0)
              Z=(4.*ZTOP(r,c)-ZTOP(r+1,c)-ZTOP(r,c+1))/2.;
            else if(r==Nrow)
              Z=(4.*ZTOP(r-1,c)-ZTOP(r-2,c)-ZTOP(r-1,c+1))/2.;
            else
              Z=(3.*ZTOP(r-1,c)-ZTOP(r-1,c+1)+3.*ZTOP(r,c)-ZTOP(r,c+1))/4.;
            }
          else if(c==Ncol)
            {
            if(r==0)
              Z=(4.*ZTOP(r,c-1)-ZTOP(r+1,c-1)-ZTOP(r,c-2))/2.;
            else if(r==Nrow)
              Z=(4.*ZTOP(r-1,c-1)-ZTOP(r-2,c-1)-ZTOP(r-1,c-2))/2.;
            else
              Z=(3.*ZTOP(r-1,c-1)-ZTOP(r-1,c-2)+3.*ZTOP(r,c-1)-ZTOP(r,c-2))/4.;
            }
          else
            {
            if(r==0)
              Z=(3.*ZTOP(r,c-1)-ZTOP(r+1,c-1)+3.*ZTOP(r,c)-ZTOP(r+1,c))/4.;
            else if(r==Nrow)
              Z=(3.*ZTOP(r-1,c-1)-ZTOP(r-2,c-1)+3.*ZTOP(r-1,c)-ZTOP(r-2,c))/4.;
            else
              Z=(ZTOP(r-1,c-1)+ZTOP(r-1,c)+ZTOP(r,c-1)+ZTOP(r,c))/4.;
            }
          Zmax=max(Zmax,Z);
          Zmin=min(Zmin,Z);
          Node[node(x,y,z)].Z=Z;
          }
        }
#undef ZTOP
      if(l<Nlay-1)
        break;
      Zswp=Zbot;
      Zbot=Ztop;
      Ztop=Zswp;
      }
    free(Zbot);
    free(Ztop);
    }
  printf("  %lGZ%lG\n",Zmin,Zmax);
  Xsiz=Xmax-Xmin;
  Ysiz=Ymax-Ymin;
  Zsiz=Zmax-Zmin;
  if(Xsiz<=0)
    Abort(__LINE__,"domain is degenerate in X");
  if(Ysiz<=0)
    Abort(__LINE__,"domain is degenerate in Y");
  if(Zsiz<=0)
    Abort(__LINE__,"domain is degenerate in Z");
  for(n=e=0;e<Ne;e++)
    {
    for(i=0;i<8;i++,n++)
      {
      Elem[e].X+=Node[El[n]].X;
      Elem[e].Y+=Node[El[n]].Y;
      Elem[e].Z+=Node[El[n]].Z;
      }
    Elem[e].X/=8.;
    Elem[e].Y/=8.;
    Elem[e].Z/=8.;
    }
  }

int rdbin(int code,void*bufr,int bytes,FILE*fp)
  {
  if(fread(bufr,1,bytes,fp)!=bytes)
    Abort(code,"unexpected end of binary file");
  return(bytes);
  }

float*ReadBinary(FILE*fp)
  {
  int i;
  float*F;
  rdbin(__LINE__,&i,sizeof(int),fp);
  if(i!=1)
    Abort(__LINE__,"KSTP must be 1");
  rdbin(__LINE__,&i,sizeof(int),fp);
  if(i!=1)
    Abort(__LINE__,"KPER must be 1");
  rdbin(__LINE__,cbuf,16,fp);
  rdbin(__LINE__,&i,sizeof(int),fp);
  if(i!=Ncol)
    Abort(__LINE__,"Ncol conflict");
  rdbin(__LINE__,&i,sizeof(int),fp);
  if(i!=Nrow)
    Abort(__LINE__,"Nrow conflict");
  rdbin(__LINE__,&i,sizeof(int),fp);
  if(i!=Nlay)
    Abort(__LINE__,"Nlay conflict");
  F=allocate(__LINE__,Ne,sizeof(float));
  rdbin(__LINE__,F,Ne*sizeof(float),fp);
  return(F);
  }

void ReadBinaryFlowFile()
  {
  char fname[13];
  int c,e,i,l,r,x,y,z;
  float*UVW;
  FILE*fp;
  strcpy(fname,prefix);
  strcat(fname,".CBB");
  printf("binary flow file: %s\n",fname);
  if((fp=fopen(fname,"rb"))==NULL)
    Abort(__LINE__,"can't open essential input file");
  free(ReadBinary(fp));
  UVW=ReadBinary(fp);
  for(l=i=0;l<Nlay;l++)
    {
    z=Nlay-l-1;
    for(r=0;r<Nrow;r++)
      {
      y=Nrow-r-1;
      for(c=0;c<Ncol;c++,i++)
        {
        x=c;
        e=elem(x,y,z);
        if(Elem[e].active)
          Elem[e].U=UVW[i]/dR[r]/dL[e]/Elem[e].F;
        else if(UVW[i]!=0.)
          {
          sprintf(cbuf,"inactive cell col=%i, row=%i, layer=%i\nis inactive yet has a non-zero X velocity",c+1,r+1,l+1);
          Abort(__LINE__,cbuf);
          }
        }
      }
    }
  free(UVW);
  UVW=ReadBinary(fp);
  for(l=i=0;l<Nlay;l++)
    {
    z=Nlay-l-1;
    for(r=0;r<Nrow;r++)
      {
      y=Nrow-r-1;
      for(c=0;c<Ncol;c++,i++)
        {
        x=c;
        e=elem(x,y,z);
        if(Elem[e].active)
          Elem[e].V=-UVW[i]/dC[c]/dL[e]/Elem[e].F;
        else if(UVW[i]!=0.)
          {
          sprintf(cbuf,"inactive cell col=%i, row=%i, layer=%i\nis inactive yet has a non-zero Y velocity",c+1,r+1,l+1);
          Abort(__LINE__,cbuf);
          }
        }
      }
    }
  free(UVW);
  UVW=ReadBinary(fp);
  for(l=i=0;l<Nlay;l++)
    {
    z=Nlay-l-1;
    for(r=0;r<Nrow;r++)
      {
      y=Nrow-r-1;
      for(c=0;c<Ncol;c++,i++)
        {
        x=c;
        e=elem(x,y,z);
        if(Elem[e].active)
          Elem[e].W=-UVW[i]/dC[c]/dR[r]/Elem[e].F;
        else if(UVW[i]!=0.)
          {
          sprintf(cbuf,"inactive cell col=%i, row=%i, layer=%i\nis inactive yet has a non-zero Z velocity",c+1,r+1,l+1);
          Abort(__LINE__,cbuf);
          }
        }
      }
    }
  free(UVW);
  fclose(fp);
  Umin=Umax=Elem[0].U;
  Vmin=Vmax=Elem[0].V;
  Wmin=Wmax=Elem[0].W;
  for(z=e=0;z<Nlay;z++)
    {
    for(y=0;y<Nrow;y++)
      {
      for(x=0;x<Ncol;x++,e++)
        {
        Umin=fmin(Umin,Elem[e].U);
        Umax=fmax(Umax,Elem[e].U);
        Vmin=fmin(Vmin,Elem[e].V);
        Vmax=fmax(Vmax,Elem[e].V);
        Wmin=fmin(Wmin,Elem[e].W);
        Wmax=fmax(Wmax,Elem[e].W);
        }
      }
    }
  printf("  %lGU%lG\n",Umin,Umax);
  printf("  %lGV%lG\n",Vmin,Vmax);
  printf("  %lGW%lG\n",Wmin,Wmax);
  }

void ReadPropertyFile()
  {
  char fname[13];
  int c,e,i,l,r,x,y,z;
  double Dx,Dy,Dz,F;
  FILE*fp;
  strcpy(fname,prefix);
  strcat(fname,".ELP");
  printf("element property file: %s\n",fname);
  if((fp=fopen(fname,"rt"))==NULL)
    Abort(__LINE__,"can't open essential input file");
  if(!fgets(cbuf,sizeof(cbuf),fp))
    Abort(__LINE__,"unexpected end of file");
  if(sscanf(cbuf,"%i",&e)!=1)
    Abort(__LINE__,"scan error on number of elements");
  if(e!=Ne)
    Abort(__LINE__,"number of elements don't match %i?%i",Ne,e);
  Fmin=Dxmn=Dymn=Dzmn=DBL_MAX;
  Fmax=Dxmx=Dymx=Dzmx=-DBL_MAX;
  for(l=i=0;l<Nlay;l++)
    {
    z=Nlay-l-1;
    for(r=0;r<Nrow;r++)
      {
      y=Nrow-r-1;
      for(c=0;c<Ncol;c++,i++)
        {
        x=c;
        e=elem(x,y,z);
        if(!fgets(cbuf,sizeof(cbuf),fp))
          Abort(__LINE__,"unexpected end of file");
        if(sscanf(cbuf,"%lG %lG %lG %lG",&F,&Dx,&Dy,&Dz)!=4)
          Abort(__LINE__,"scan error on F Dx Dy Dz for element %i",e+1);
        if(Elem[e].active)
          {
          if(F<=0.)
            Abort(__LINE__,"porosity<=0 for element %i",e+1);
          Elem[e].F=F;
          Elem[e].Dx=Dx;
          Elem[e].Dy=Dy;
          Elem[e].Dz=Dz;
          Fmin=fmin(Fmin,F);
          Fmax=fmax(Fmax,F);
          Dxmn=fmin(Dxmn,Dx);
          Dxmx=fmax(Dxmx,Dx);
          Dymn=fmin(Dymn,Dy);
          Dymx=fmax(Dymx,Dy);
          Dzmn=fmin(Dzmn,Dz);
          Dzmx=fmax(Dzmx,Dz);
          }
        }
      }
    }
  printf("  %lGF%lG\n",Fmin,Fmax);
  printf("  %lGDx%lG\n",Dxmn,Dxmx);
  printf("  %lGDy%lG\n",Dymn,Dymx);
  printf("  %lGDz%lG\n",Dzmn,Dzmx);
  }

void ReadConcentrations()
  {
  char fname[13];
  int e,i,n,x,y,z;
  double C;
  FILE*fp;
  strcpy(fname,prefix);
  strcat(fname,".ICC");
  printf("initial concentrations: %s\n",fname);
  if((fp=fopen(fname,"rt"))==NULL)
    Abort(__LINE__,"can't open essential input file");
  for(z=Nlay-1;z>=0;z--)        /* read top-to-bottom */
    {
    if(!fgets(cbuf,sizeof(cbuf),fp))
      Abort(__LINE__,"unexpected end of file");
    if(sscanf(cbuf,"%i",&i)!=1)
      Abort(__LINE__,"scan error on layer");
    if(i!=Nlay-z)
      Abort(__LINE__,"expected layer %i but found %i",Nlay-z,i);
    if(!fgets(cbuf,sizeof(cbuf),fp))
      Abort(__LINE__,"unexpected end of file");
    for(i=0,y=Nrow-1;y>=0;y--)  /* read north-to-south */
      {
      for(x=0;x<Ncol;x++,i++)   /* read west-to-east */
        {
        while(cbuf[i]==' ')
          i++;
        if(cbuf[i]==0)
          {
          if(!fgets(cbuf,sizeof(cbuf),fp))
            Abort(__LINE__,"unexpected end of file");
          i=0;
          }
        if(sscanf(cbuf+i,"%lf%n",&C,&n)!=1)
          Abort(__LINE__,"scan error on concentration");
        i+=n;
        e=elem(x,y,z);
        if(Elem[e].active)
          Node[El[e*8]].C=C;
        }
      }
    }
  fclose(fp);
  }

void CreateVelocityFile2D()
  {
  char fname[13];
  int e,n,x,y,z;
  double U,V;
  FILE*fp;
  strcpy(fname,prefix);
  strcat(fname,".V2D");
  printf("vertically-averaged velocities: %s\n",fname);
  if((fp=fopen(fname,"wt"))==NULL)
    Abort(__LINE__,"can't create output file");
  for(y=0;y<Nrow;y++)
    {
    for(x=0;x<Ncol;x++)
      {
      for(U=V=n=z=0;z<Nlay;z++)
        {
        e=elem(x,y,z);
        if(Elem[e].active)
          {
          U+=Elem[e].U;
          V+=Elem[e].V;
          n++;
          }
        }
      if(n)
        fprintf(fp,"%lG %lG %lG %lG\n",Elem[e].X,Elem[e].Y,U/n,V/n);
      }
    }
  fclose(fp);
  }

void AdvanceSolution(double t,double dt)
  {
  int e,ic,ie,il,in,is,iu,iw,x,y,z;
  double Ca,Cb,Cc,Cd,Ce,Cf,Cg,Ch,Cl,Cn,Cs,Cu,Cw,decay,dX,Dx,dY,Dy,dZ,Dz,U,V,W;
  double d2CdX2,d2CdY2,d2CdZ2,dCdXa,dCdXb,dCdYd,dCdYf,dCdZg,dCdZh,UdCdX,VdCdY,WdCdZ;
  decay=exp(-(t+dt)/tau)/exp(-t/tau);
  for(z=1;z<Nlay;z++)
    {
    for(y=1;y<Nrow;y++)
      {
      for(x=1;x<Ncol;x++)
        {
        ic=node(x,y,z);
        if(Node[ic].active)
          Node[ic].C*=decay;
        }
      }
    }
  for(z=1;z<Nlay;z++)
    {
    for(y=1;y<Nrow;y++)
      {
      dY=dR[Nrow-1-y];
      for(x=1;x<Ncol;x++)
        {
        dX=dC[x-1];
        dZ=dL[elem(x-1,Nrow-1-y,Nlay-1-z)];
        ic=node(x,y,z);
        if(!Node[ic].active)
          continue;
        ie=node(x+1,y,z);
        iw=node(x-1,y,z);
        in=node(x,y+1,z);
        is=node(x,y-1,z);
        iu=node(x,y,z+1);
        il=node(x,y,z-1);
        Cc=Node[ic].C;
        Ce=Node[ie].C;
        Cn=Node[in].C;
        Cs=Node[is].C;
        Cw=Node[iw].C;
        Cu=Node[iu].C;
        Cl=Node[il].C;
        Ca=(Cw+Cc)/2.;
        Cb=(Ce+Cc)/2.;
        Cd=(Cn+Cc)/2.;
        Cf=(Cs+Cc)/2.;
        Cg=(Cu+Cc)/2.;
        Ch=(Cl+Cc)/2.;
        dCdXa=(Cc-Cw)/dX;
        dCdXb=(Ce-Cc)/dX;
        dCdYd=(Cn-Cc)/dY;
        dCdYf=(Cc-Cs)/dY;
        dCdZg=(Cu-Cc)/dZ;
        dCdZh=(Cc-Cl)/dZ;
        d2CdX2=(dCdXb-dCdXa)/dX;
        d2CdY2=(dCdYd-dCdYf)/dY;
        d2CdZ2=(dCdZg-dCdZh)/dZ;
        e=elem(x,y,z);
        Dx=Elem[e].Dx;
        Dy=Elem[e].Dy;
        Dz=Elem[e].Dz;
        U=Elem[e].U;
        V=Elem[e].V;
        W=Elem[e].W;
        if(U>0.)
          UdCdX=U*(Cc-Cw)/dX;
        else
          UdCdX=U*(Ce-Cc)/dX;
        if(V>0.)
          VdCdY=V*(Cc-Cs)/dY;
        else
          VdCdY=V*(Cn-Cc)/dY;
        if(W>0.)
          WdCdZ=W*(Cc-Cl)/dZ;
        else
          WdCdZ=W*(Cu-Cc)/dZ;
        dCdt[ic]=Dx*d2CdX2+Dy*d2CdY2+Dz*d2CdZ2-UdCdX-VdCdY-WdCdZ;
        }
      }
    }
  for(z=1;z<Nlay;z++)
    {
    for(y=1;y<Nrow;y++)
      {
      for(x=1;x<Ncol;x++)
        {
        ic=node(x,y,z);
        if(!Node[ic].active)
          continue;
        Node[ic].C+=dt*dCdt[ic];
        }
      }
    }
  }

double MaximumConcentration()
  {
  int n;
  double C;
  for(C=n=0;n<Nn;n++)
    if(Node[n].active)
      if(Node[n].C>C)
        C=Node[n].C;
  return(C);
  }

void FileResults(char*name)
  {
  int n,x,y,z;
  double C,Cm,logCm;
  FILE*fp;
  printf("%s\n",name);
  if((fp=fopen(name,"wt"))==NULL)
    Abort(__LINE__,"can't create output file");
  fprintf(fp,"%i\n",Ncol+1);
  for(x=0;x<=Ncol;x++)
    fprintf(fp,"%lG\n",Node[node(x,0,0)].X);
  fprintf(fp,"%i\n",Nrow+1);
  for(y=0;y<=Nrow;y++)
    fprintf(fp,"%lG\n",Node[node(0,y,0)].Y);
  fprintf(fp,"%i\n",Nlay+1);
  for(z=0;z<=Nlay;z++)
    fprintf(fp,"%lG\n",Node[node(0,0,z)].Z);
  fprintf(fp,"%i\n",(Ncol+1)*(Nrow+1)*(Nlay+1));
  Cm=Co/1E6;
  logCm=log(Cm);
  for(z=0;z<=Nlay;z++)
    {
    for(y=0;y<=Nrow;y++)
      {
      for(x=0;x<=Ncol;x++)
        {
        n=node(x,y,z);
        if(Node[n].active)
          {
          C=Node[n].C;
          if(C<=Cm)
            fprintf(fp,"%lG\n",logCm);
          else
            fprintf(fp,"%lG\n",log(C));
          }
        else
          fprintf(fp,"-1E99\n");
        }
      }
    }
  fclose(fp);
  }

double DetermineTimeStep()
  {
  int x,y,z;
  double atx,aty,atz,dt,dtx,dty,dtz,dX,dY,dZ;
  printf("determining time step\n");
  dX=dC[0];
  for(x=1;x<Ncol;x++)
    if(dC[x]<dX)
      dX=dC[x];
  dY=dR[0];
  for(y=1;y<Nrow;y++)
    if(dR[y]<dY)
      dY=dR[y];
  dZ=dL[0];
  for(z=1;z<Ne;z++)
    if(dL[z]<dZ)
      dZ=dL[z];
  printf("  X=%lG, Y=%lG, Z=%lG\n",dX,dY,dZ);
  dtx=dX*dX/fmin(fabs(Dxmn),fabs(Dxmx))/2.;
  dty=dY*dY/fmin(fabs(Dymn),fabs(Dymx))/2.;
  dtz=dZ*dZ/fmin(fabs(Dzmn),fabs(Dzmx))/2.;
  printf("  X/Dx/2=%lG, Y/Dy/2=%lG, Z/Dz/2=%lG\n",dtx,dty,dtz);
  atx=dX/fmax(fabs(Umin),fabs(Umax))/6.;
  aty=dY/fmax(fabs(Vmin),fabs(Vmax))/6.;
  atz=dZ/fmax(fabs(Wmin),fabs(Wmax))/6.;
  printf("  X/U/6=%lG, Y/V/6=%lG, Z/W/6=%lG\n",atx,aty,atz);
  dt=1./(1./dtx+1./dty+1./dtz+1./atx+1./aty+1./atz);
  printf("  t=%lG\n",dt);
  return(dt);
  }

void CheckPrefix(int argc,char**argv)
  {
  if(argc<2)
    Abort(__LINE__,"no input specified... try\nctmod wine");
  if(argc>2)
    Abort(__LINE__,"too many parameters specified... try\nctmod wine");
  printf("prefix: %s\n",argv[1]);
  if(strlen(argv[1])<3)
    Abort(__LINE__,"prefix must be at least 3 characters");
  if(strlen(argv[1])>5)
    Abort(__LINE__,"prefix must be no more than 5 characters");
  if(strchr(argv[1],' '))
    Abort(__LINE__,"prefix must not contain blanks");
  if(strchr(argv[1],'.'))
    Abort(__LINE__,"prefix must not contain a dot or have an extension");
  strcpy(prefix,argv[1]);
  if(strspn(prefix,"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_0123456789")!=strlen(prefix))
    Abort(__LINE__,"prefix contains illegal character (must be A-Za-z_0-9)");
  _strupr(prefix);
  }

int main(int argc,char**argv,char**envp)
  {
  char fname[13];
  int i,j,k=0,n;
  double Cx,dt,t;
  printf("3D Transient Diffusion + Dispersion + Advection\n");
  printf("reading MODFLOW input and output files\n");
  CheckPrefix(argc,argv);
  ReadBasicInputOptionFile();
  ReadBlockCenteredFile();
  ReadPropertyFile();
  ReadBinaryFlowFile();
  CreateVelocityFile2D();
  dt=DetermineTimeStep();
  restart:
  ReadConcentrations();
  Co=MaximumConcentration();
  printf("0 years C%.0lf ",Co);
  sprintf(fname,"%s000.TB3",prefix);
  FileResults(fname);
  n=(int)(5./dt+0.5);
  for(t=j=i=0;t<1800.;i++)
    {
    AdvanceSolution(t,dt);
    Cx=MaximumConcentration();
    if(Cx>Co*2.)
      {
      printf("\rCx=%lG\ntime step is too large - solution is unstable\n",Cx);
      if(++k>16)
        break;
      dt/=2.;
      printf("restart with t=%lG\n",dt);
      goto restart;
      }
    t+=dt;
    printf("%.3lf years C%.0lf ",t,Cx);
    if((i+1)%n==0)
      {
      sprintf(fname,"%s%03d.TB3",prefix,++j);
      FileResults(fname);
      if(Cx<Co/10.)
        break;
      }
    else
      printf("\r");
    }
  return(0);
  }
