/* This is a transient 2D numerical model for contaminant transport
   that uses finite differences for the spatial dimensions and
   forward Euler's method for the temporal. It includes diffusion,
   dispersion, and advection. The results are painted into a bitmap
   having a one-to-one node-to-pixel correspondence.

   This code was written by Dudley J. Benton
*/

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "..\\12x08.h"
#define Font C12x08

double Ax;         /* maximum dispersion coefficient [cm/sec] */
double Ch=300.;    /* highest concentration [ppm] */
double Cl=0.03;    /* lowest concentration [ppm] */
double Co=300.;    /* initial concentration [ppm] */
double Cx;         /* maximum concentration [ppm] */
double Dh=0.0006;  /* high diffusion coefficient [cm/sec] */
double Dl=0.0001;  /* low  diffusion coefficient [cm/sec] */
double Dm=0.0003;  /* mid  diffusion coefficient [cm/sec] */
double Do= 80.;    /* span of initial contamination [m] */
double dX=1.;      /* X grid spacing [m] */
double dY=1.;      /* Y grid spacing [m] */
double UV=0.014142;/* nominal velocity [m/day] */
double Ux;         /* maximum velocity in X direction [m/day] */
double Vx;         /* maximum velocity in Y direction [m/day] */
double*Ah;         /* X dispersion [cm/sec] */
double*Av;         /* Y dispersion [cm/sec] */
double*C;          /* concentration array [ppm] */
double*D;          /* diffusion coefficient array [cm/sec] */
double*dCdt;       /* working space/supplemental array */
double*F;          /* porosity [unitless] */
double*U;          /* X velocity [m/day] */
double*V;          /* Y velocity [m/day] */

#define day_per_yr 365.25
#define sec_per_yr (3600.*24.*365.25)
#define cm_per_m     100.

#define Nx 640
#define Ny 640

struct{
  BITMAPFILEHEADER bf;
  BITMAPINFOHEADER bi;
  BYTE*bits;
  int wide;
  }bm;

DWORD palette[]={   0x000000,0xFFFFFF,0x7F7F7F,0x0000FF,0x0008FF,0x0010FF,
  0x0018FF,0x0020FF,0x0028FF,0x0030FF,0x0038FF,0x0040FF,0x0048FF,0x0050FF,
  0x0058FF,0x0060FF,0x0068FF,0x0070FF,0x0078FF,0x0080FF,0x0088FF,0x0090FF,
  0x0098FF,0x00A0FF,0x00A8FF,0x00B0FF,0x00B8FF,0x00C0FF,0x00C8FF,0x00D0FF,
  0x00D8FF,0x00E0FF,0x00E8FF,0x00F0FF,0x00F8FF,0x00FFF8,0x00FFF0,0x00FFE8,
  0x00FFE0,0x00FFD8,0x00FFD0,0x00FFC8,0x00FFC0,0x00FFB8,0x00FFB0,0x00FFA8,
  0x00FFA0,0x00FF98,0x00FF90,0x00FF88,0x00FF80,0x00FF78,0x00FF70,0x00FF68,
  0x00FF60,0x00FF58,0x00FF50,0x00FF48,0x00FF40,0x00FF38,0x00FF30,0x00FF28,
  0x00FF20,0x00FF18,0x00FF10,0x00FF08,0x00FF00,0x08FF00,0x10FF00,0x18FF00,
  0x20FF00,0x28FF00,0x30FF00,0x38FF00,0x40FF00,0x48FF00,0x50FF00,0x58FF00,
  0x60FF00,0x68FF00,0x70FF00,0x78FF00,0x80FF00,0x88FF00,0x90FF00,0x98FF00,
  0xA0FF00,0xA8FF00,0xB0FF00,0xB8FF00,0xC0FF00,0xC8FF00,0xD0FF00,0xD8FF00,
  0xE0FF00,0xE8FF00,0xF0FF00,0xF8FF00,0xFFFF00,0xFFF800,0xFFF000,0xFFE800,
  0xFFE000,0xFFD800,0xFFD000,0xFFC800,0xFFC000,0xFFB800,0xFFB000,0xFFA800,
  0xFFA000,0xFF9800,0xFF9000,0xFF8800,0xFF8000,0xFF7800,0xFF7000,0xFF6800,
  0xFF6000,0xFF5800,0xFF5000,0xFF4800,0xFF4000,0xFF3800,0xFF3000,0xFF2800,
  0xFF2000,0xFF1800,0xFF1000,0xFF0800,0xFF0000};

#define black 0x00
#define white 0x01
#define gray  0x02
#define blue  0x03
#define red   0x82

void EmbedText(int col,int row,char*string,BYTE fore,BYTE back,int transparent)
  {
  BYTE*font,mask,s;
  int h,i,j,w;
  while(*string)
    {
    s=*string;
    string++;
    i=Font[0]*(DWORD)Font[1]*(DWORD)s;
    font=Font+2+i/8;
    mask=0x80>>(i%8);
    for(h=0;h<Font[0];h++)
      {
      j=bm.wide*(row-h)+col;
      for(w=0;w<Font[1];w++,j++)
        {
        if(mask&*font)
          bm.bits[j]=fore;
        else if(!transparent)
          bm.bits[j]=back;
        mask>>=1;
        if(mask==0)
          {
          mask=0x80;
          font++;
          }
        }
      }
    col+=Font[1];
    }
  }

double fmin(double x,double y)
  {
  if(x<y)
    return(x);
  return(y);
  }

double fmax(double x,double y)
  {
  if(x>y)
    return(x);
  return(y);
  }

void Abort(int code,char*format,...)
  {
  va_list arg_marker;
  fprintf(stderr,"Fatal Error Code %i\n",code);
  va_start(arg_marker,format);
  vfprintf(stderr,format,arg_marker);
  fprintf(stderr,"\n");
  _fcloseall();
  exit(1);
  }

void InitializeBitmap()
  {
  printf("bitmap: %ix%i\n",Nx,Ny);
  if(Nx%4)
    Abort(__LINE__,"Nx must be an even multiple of 4");
  bm.bi.biSize=sizeof(BITMAPINFOHEADER);
  bm.bi.biPlanes=1;
  bm.bi.biBitCount=8;
  bm.bi.biWidth=Nx;
  bm.bi.biHeight=Ny;
  bm.bi.biClrUsed=sizeof(palette)/sizeof(DWORD);
  bm.wide=4*((bm.bi.biWidth*bm.bi.biBitCount+31)/32);
  bm.bi.biSizeImage=bm.wide*bm.bi.biHeight;
  if((bm.bits=calloc(bm.bi.biSizeImage,sizeof(BYTE)))==NULL)
    Abort(__LINE__,"can't allocate memory for bitmap");
  bm.bf.bfType=0x4D42;
  bm.bf.bfOffBits=sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+sizeof(palette);
  bm.bf.bfSize=bm.bf.bfOffBits+bm.bi.biSizeImage;
  }

void InitializePorosity()
  {
  int i,x,y;
  printf("pororisty: %ix%i\n",Nx,Ny);
  if((F=calloc(Nx*Ny,sizeof(double)))==NULL)
    Abort(__LINE__,"can't allocate memory for diffusion coefficient");
  for(i=y=0;y<Ny;y++)
    for(x=0;x<Nx;x++,i++)
      F[i]=0.3;
  }

double hiD(double X)
  {
  return((((7.09055811489E-8*X-6.54449479371E-5)*X
    +0.022490660023)*X-4.36026882252)*X+410.588712226);
  }

double loD(double Y)
  {
  return((((-1.57591948391E-9*Y+5.632933722E-7)*Y
    -0.00107402852929)*Y-0.669261054953)*Y-118.487925182);
  }

void InitializeDiffusion()
  {
  int i,x,y;
  double X,Y;
  printf("diffusion: %ix%i\n",Nx,Ny);
  Dl*=sec_per_yr/cm_per_m/cm_per_m;
  Dm*=sec_per_yr/cm_per_m/cm_per_m;
  Dh*=sec_per_yr/cm_per_m/cm_per_m;
  if((D=calloc(Nx*Ny,sizeof(double)))==NULL)
    Abort(__LINE__,"can't allocate memory for diffusion coefficient");
  for(i=y=0;y<Ny;y++)
    {
    Y=(Ny/2-y)*dY;
    for(x=0;x<Nx;x++,i++)
      {
      X=(Nx/2-x)*dX;
      if(Y>hiD(X))
        D[i]=Dl;
      else if(X<loD(Y))
        D[i]=Dh;
      else
        D[i]=Dm;
      }
    }
  }

double aX(double X,double Y)
  {
  return((0.01967332026+(0.000005564165715-1.718383973E-9*Y)*Y
    +(1.063303373E-8*Y-0.00001510726646-3.128773849E-9*X)*X)/4.);
  }

double aY(double X,double Y)
  {
  return((0.006326680436+(-0.000005564163394+1.718357281E-9*Y)*Y
    +(-1.063306988E-8*Y+0.00001510727151+3.128765211E-9*X)*X)/4.);
  }

void InitializeDispersion()
  {
  int i,x,y;
  double X,Y;
  printf("dispersion: %ix%i\n",Nx,Ny);
  if((Ah=calloc(Nx*Ny,sizeof(double)))==NULL)
    Abort(__LINE__,"can't allocate memory for dispersion coefficient");
  if((Av=calloc(Nx*Ny,sizeof(double)))==NULL)
    Abort(__LINE__,"can't allocate memory for dispersion coefficient");
  Ax=0.;
  for(i=y=0;y<Ny;y++)
    {
    Y=(Ny/2-y)*dY;
    for(x=0;x<Nx;x++,i++)
      {
      X=(Nx/2-x)*dX;
      Ah[i]=aX(X,Y)*sec_per_yr/cm_per_m/cm_per_m;
      Av[i]=aY(X,Y)*sec_per_yr/cm_per_m/cm_per_m;
      if(Ah[i]>Ax)
        Ax=Ah[i];
      if(Av[i]>Ax)
        Ax=Av[i];
      }
    }
  }

double ang(double X,double Y)
  {
  return(-0.196349541+(0.0004601942365-4.793689941E-7*Y)*Y
    +(0.000001438106987*Y-0.000715857701+9.251858539E-22*X)*X);
  }

void InitializeVelocity()
  {
  int i,x,y;
  double a,X,Y;
  printf("velocity: %ix%i\n",Nx,Ny);
  if((U=calloc(Nx*Ny,sizeof(double)))==NULL)
    Abort(__LINE__,"can't allocate memory for velocity");
  if((V=calloc(Nx*Ny,sizeof(double)))==NULL)
    Abort(__LINE__,"can't allocate memory for velocity");
  Ux=Vx=0.;
  for(i=y=0;y<Ny;y++)
    {
    Y=(Ny/2-y)*dY;
    for(x=0;x<Nx;x++,i++)
      {
      X=(Nx/2-x)*dX;
      a=ang(X,Y);
      U[i]=UV*cos(a)*day_per_yr;
      V[i]=UV*sin(a)*day_per_yr;
      Ux=fmax(Ux,fabs(U[i]));
      Vx=fmax(Vx,fabs(U[i]));
      }
    }
  }

void InitializeConcentration()
  {
  int i,x,y;
  double X,Y;
  printf("field: %ix%i\n",Nx,Ny);
  if((C=calloc(Nx*Ny,sizeof(double)))==NULL)
    Abort(__LINE__,"can't allocate memory for concentration");
  if((dCdt=calloc(Nx*Ny,sizeof(double)))==NULL)
    Abort(__LINE__,"can't allocate memory for working array");
  for(i=y=0;y<Ny;y++)
    {
    Y=(Ny/2-y)*dY;
    for(x=0;x<Nx;x++,i++)
      {
      X=(Nx/2-x)*dX;
      C[i]=0.;
      if(fabs(Y)<=Do/2.)
        if(fabs(X)<Do/2.-Y)
          C[i]=Co;
      }
    }
  }

void AdvanceSolution(double dt)
  {
  int ic,ie,in,is,iw,x,y;
  double Aa,Ab,Ad,Af,Ca,Cb,Cc,Cd,Ce,Cf,Cn,Cs,Cw,Da,Db,Dd,Df,Fc,Fe,Fn,Fs,Fw,Uc,Vc;
  double d2ADFCdX2,d2ADFCdY2,dFCdXa,dFCdXb,dFCdYd,dFCdYf,UdFCdX,VdFCdY;
  for(y=1;y<Ny-1;y++)
    {
    for(x=1;x<Nx-1;x++)           /* interior points */
      {
      ic=Nx*y+x;
      ie=ic+1;
      iw=ic-1;
      in=ic+Nx;
      is=ic-Nx;
      Cc=C[ic];
      Ce=C[ie];
      Cn=C[in];
      Cs=C[is];
      Cw=C[iw];
      Fc=F[ic];
      Fe=F[ie];
      Fn=F[in];
      Fs=F[is];
      Fw=F[iw];
      Uc=U[ic];
      Vc=V[ic];
      Aa=(Ah[iw]+Ah[ic])/2.;
      Ab=(Ah[ie]+Ah[ic])/2.;
      Ad=(Av[in]+Av[ic])/2.;
      Af=(Av[is]+Av[ic])/2.;
      Ca=(C[iw]+C[ic])/2.;
      Cb=(C[ie]+C[ic])/2.;
      Cd=(C[in]+C[ic])/2.;
      Cf=(C[is]+C[ic])/2.;
      Da=(D[iw]+D[ic])/2.;
      Db=(D[ie]+D[ic])/2.;
      Dd=(D[in]+D[ic])/2.;
      Df=(D[is]+D[ic])/2.;
      dFCdXa=(Fc*Cc-Fw*Cw)/dX;
      dFCdXb=(Fe*Ce-Fc*Cc)/dX;
      dFCdYd=(Fn*Cn-Fc*Cc)/dY;
      dFCdYf=(Fc*Cc-Fs*Cs)/dY;
      d2ADFCdX2=((Ab+Db)*dFCdXb-(Aa+Da)*dFCdXa)/dX;
      d2ADFCdY2=((Ad+Dd)*dFCdYd-(Af+Df)*dFCdYf)/dY;
      if(Uc>0.)
        UdFCdX=Uc*(Fc*Cc-Fw*Cw)/dX;
      else
        UdFCdX=Uc*(Fe*Ce-Fc*Cc)/dX;
      if(Vc>0.)
        VdFCdY=Vc*(Fc*Cc-Fs*Cs)/dY;
      else
        VdFCdY=Vc*(Fn*Cn-Fc*Cc)/dY;
      dCdt[ic]=(d2ADFCdX2+d2ADFCdY2-UdFCdX-VdFCdY)/Fc;
      }
    }
  Cx=0.;
  for(y=1;y<Ny-1;y++)           /* interior points */
    {
    for(x=1;x<Nx-1;x++)
      {
      ic=Nx*y+x;
      C[ic]+=dt*dCdt[ic];
      if(C[ic]>Cx)
        Cx=C[ic];
      }
    }
  for(y=0,x=1;x<Nx-1;x++)       /* bottom side */
    C[Nx*y+x]=C[Nx*(y+1)+x];
  for(y=Ny-1,x=1;x<Nx-1;x++)    /* top side */
    C[Nx*y+x]=C[Nx*(y-1)+x];
  for(x=0,y=1;y<Ny-1;y++)       /* left side */
    C[Nx*y+x]=C[Nx*y+x+1];
  for(x=Nx-1,y=1;y<Ny-1;y++)    /* right side */
    C[Nx*y+x]=C[Nx*y+x+1];
  x=y=0;                        /* bottom left corner */
  C[Nx*y+x]=(C[Nx*y+x+1]+C[Nx*(y+1)+x]+C[Nx*(y+1)+x+1])/3.;
  x=Nx-1;y=0;                   /* bottom right corner */
  C[Nx*y+x]=(C[Nx*y+x-1]+C[Nx*(y+1)+x]+C[Nx*(y+1)+x-1])/3.;
  x=0;y=Ny-1;                   /* top left corner */
  C[Nx*y+x]=(C[Nx*y+x+1]+C[Nx*(y-1)+x]+C[Nx*(y-1)+x+1])/3.;
  x=Nx-1;y=Ny-1;                /* top right corner */
  C[Nx*y+x]=(C[Nx*y+x-1]+C[Nx*(y-1)+x]+C[Nx*(y-1)+x-1])/3.;
  }

void PaintConcentration()
  {
  int c,i,x,y;
  double hC,iC,lC;
  hC=log(Ch);
  lC=log(Cl);
  for(i=y=0;y<Ny;y++)
    {
    for(x=0;x<Nx;x++,i++)
      {
      if(C[i]<Cl)
        bm.bits[i]=gray;
      else if(C[i]>=Ch)
        bm.bits[i]=red;
      else
        {
        iC=log(C[i]);
        c=blue+(int)((iC-lC)*(red-blue)/(hC-lC));
        bm.bits[i]=(BYTE)c;
        }
      }
    }
  }

void WriteBitmap(char*name)
  {
  FILE*fp;
  printf("%s",name);
  if((fp=fopen(name,"wb"))==NULL)
    Abort(__LINE__,"can't create output file");
  fwrite(&bm.bf,sizeof(BITMAPFILEHEADER),1,fp);
  fwrite(&bm.bi,sizeof(BITMAPINFOHEADER),1,fp);
  fwrite(palette,sizeof(palette)/sizeof(DWORD),sizeof(DWORD),fp);
  fwrite(bm.bits,1,bm.bi.biSizeImage,fp);
  fclose(fp);
  }

int main(int argc,char**argv,char**envp)
  {
  char bufr[13],name[13];
  int i,j,n;
  double dt,t;
  printf("2D Transient Diffusion + Dispersion + Advection\n");
  InitializeBitmap();
  InitializePorosity();
  InitializeDiffusion();
  InitializeDispersion();
  InitializeVelocity();
  InitializeConcentration();
  printf("D%lG cm/sec\n",Dh*cm_per_m*cm_per_m/sec_per_yr);
  printf("A%lG cm/sec\n",Ax*cm_per_m*cm_per_m/sec_per_yr);
  UV=fmax(FLT_EPSILON,hypot(Ux,Vx));
  printf("U|V%lG m/day\n",UV/day_per_yr);
  dt=fmin(dX*dX/(Ax+Dh)/6.,dX/UV/6.);
  printf("t=%lG yr\n",dt);
  printf("A|Dt/X=%lG\n",dt*(Ax+Dh)/dX/dX);
  printf("U|Vt/X=%lG\n",dt*UV/dX);
  PaintConcentration();
  strcpy(bufr,"0 years");
  EmbedText(4,16,bufr,white,black,TRUE);
  printf("%s ",bufr);
  sprintf(bufr,"C%.0lf",Co);
  EmbedText(Nx-8*(strlen(bufr)+1),16,bufr,white,black,TRUE);
  WriteBitmap("PCE2D000.BMP");
  printf(" %s\n",bufr);
  n=(int)(1./dt+0.5);
  for(t=j=i=0;t<1800.;i++)
    {
    AdvanceSolution(dt);
    if(Cx>Co)
      Abort(__LINE__,"time step is too large\nsolution is unstable");
    t+=dt;
    if((i+1)%n==0)
      {
      PaintConcentration();
      sprintf(bufr,"%.0lf years",t);
      EmbedText(4,16,bufr,white,black,TRUE);
      printf("%s ",bufr);
      sprintf(bufr,"C%.0lf",Cx);
      EmbedText(Nx-8*(strlen(bufr)+1),16,bufr,white,black,TRUE);
      sprintf(name,"PCE2D%03d.BMP",++j);
      WriteBitmap(name);
      printf(" %s\n",bufr);
      if(Cx<Co/10.)
        break;
      }
    }
  return(0);
  }
