/* This is a transient 3D numerical model for contaminant transport
   that uses finite differences for the spatial dimensions and
   forward Euler's method for the temporal. Upwind differences are
   used for the advective terms. This model includes diffusion and
   dispersion (combined) and also advection. The results are written
   to an output file that is compatible with TP2, which can be used
   to display the results or convert the files to the Tecplot format.

   This code was written by Dudley J. Benton
*/

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

double Cm=0.003;/* minimum concentration [ppm] (for displaying the log) */
double Co;      /* initial concentration [ppm] */
double Cx;      /* maximum concentration [ppm] */
double dX=25.;  /* X grid spacing [m] */
double dY=25.;  /* Y grid spacing [m] */
double dZ=2.5;  /* Z grid spacing [m] */
double*C;       /* concentration array [ppm] */
double*dCdt;    /* working space/supplemental array */
double*Dx;      /* X diffusion coefficient array [cm/sec] */
double*Dy;      /* Y diffusion coefficient array [cm/sec] */
double*Dz;      /* Z diffusion coefficient array [cm/sec] */
double*F;       /* porosity [unitless] */
double*U;       /* X velocity [m/day] */
double*V;       /* Y velocity [m/day] */
double*W;       /* Z velocity [m/day] */

#define day_per_yr 365.25
#define sec_per_yr (3600.*24.*365.25)
#define cm_per_m   100.

#define Nx 201
#define Ny 201
#define Nz  21

void Abort(int code,char*format,...)
  {
  va_list arg_marker;
  fprintf(stderr,"Fatal Error Code %i\n",code);
  va_start(arg_marker,format);
  vfprintf(stderr,format,arg_marker);
  fprintf(stderr,"\n");
  _fcloseall();
  exit(1);
  }

void*allocate(int line,size_t count,size_t size)
  {
  void*ptr;
  if((ptr=calloc(count,size))==NULL)
    Abort(line,"can't allocate memory %ux%u=%u",count,size,count*size);
  return(ptr);
  }

double fmin(double x,double y)
  {
  if(x<y)
    return(x);
  return(y);
  }

double fmax(double x,double y)
  {
  if(x>y)
    return(x);
  return(y);
  }

double Minimum(double*Q,int n)
  {
  int i;
  double Qm;
  Qm=Q[0];
  for(i=0;i<n;i++)
    if(Q[i]<Qm)
      Qm=Q[i];
  return(Qm);
  }

double Maximum(double*Q,int n)
  {
  int i;
  double Qx;
  Qx=Q[0];
  for(i=0;i<n;i++)
    if(Q[i]>Qx)
      Qx=Q[i];
  return(Qx);
  }

void InitializePorosity()
  {
  int i,x,y,z;
  printf("pororisty: %ix%i\n",Nx,Ny);
  F=allocate(__LINE__,Nx*Ny*Nz,sizeof(double));
  for(i=z=0;z<Nz;z++)
    for(y=0;y<Ny;y++)
      for(x=0;x<Nx;x++,i++)
        F[i]=0.3;
  }

void InitializeDiffusion()
  {
  int i,x,y,z;
  printf("diffusion: %ix%ix%i\n",Nx,Ny,Nz);
  Dx=allocate(__LINE__,Nx*Ny*Nz,sizeof(double));
  Dy=allocate(__LINE__,Nx*Ny*Nz,sizeof(double));
  Dz=allocate(__LINE__,Nx*Ny*Nz,sizeof(double));
  for(i=z=0;z<Nz;z++)
    {
    for(y=0;y<Ny;y++)
      {
      for(x=0;x<Nx;x++,i++)
        {
        Dy[i]=0.0880000*sec_per_yr/cm_per_m/cm_per_m;
        Dy[i]=0.0008800*sec_per_yr/cm_per_m/cm_per_m;
        Dz[i]=0.0000088*sec_per_yr/cm_per_m/cm_per_m;
        }
      }
    }
  }

double ang(double X,double Y)
  {
  return(-0.196349541+(0.0004601942365-4.793689941E-7*Y)*Y
    +(0.000001438106987*Y-0.000715857701+9.251858539E-22*X)*X);
  }

void InitializeVelocity()
  {
  int i,x,y,z;
  double a,UV=0.014142,X,Y;
  printf("velocity: %ix%i\n",Nx,Ny);
  U=allocate(__LINE__,Nx*Ny*Nz,sizeof(double));
  V=allocate(__LINE__,Nx*Ny*Nz,sizeof(double));
  W=allocate(__LINE__,Nx*Ny*Nz,sizeof(double));
  for(i=z=0;z<Nz;z++)
    {
    for(y=0;y<Ny;y++)
      {
      Y=(Ny/2-y)*dY;
      for(x=0;x<Nx;x++,i++)
        {
        X=(Nx/2-x)*dX;
        a=ang(X,Y);
        U[i]=UV*cos(a)*day_per_yr;
        V[i]=UV*sin(a)*day_per_yr;
        W[i]=-0.005*UV*day_per_yr;
        }
      }
    }
  }

void InitializeConcentration()
  {
  int i,x,y,z;
  double R,X,Y,Z;
  printf("concentration: %ix%ix%i\n",Nx,Ny,Nz);
  C=allocate(__LINE__,Nx*Ny*Nz,sizeof(double));
  dCdt=allocate(__LINE__,Nx*Ny*Nz,sizeof(double));
  for(i=z=0;z<Nz;z++)
    {
    Z=(z-Nz/2)*dZ;
    for(y=0;y<Ny;y++)
      {
      Y=(y-Ny/2)*dY;
      for(x=0;x<Nx;x++,i++)
        {
        X=(x-Nx/2)*dX;
        R=hypot(X,Y);
        if(R<400.)
          if(fabs(Z)<6.)
            C[i]=300.;
        }
      }
    }
  }

int node(int x,int y,int z)
  {
  return((Ny*z+y)*Nx+x);
  }

void AdvanceSolution(double dt)
  {
  int ic,ie,il,in,is,iu,iw,x,y,z;
  double Ca,Cb,Cc,Cd,Ce,Cf,Cg,Ch,Cl,Cn,Cs,Cu,Cw,Da,Db,Dd,Df,Dg,Dh,Fc,Fe,Fl,Fn,Fs,Fu,Fw,Uc,Vc,Wc;
  double d2DFCdX2,d2DFCdY2,d2DFCdZ2,dFCdXa,dFCdXb,dFCdYd,dFCdYf,dFCdZg,dFCdZh,UdFCdX,VdFCdY,WdFCdZ;
  for(z=1;z<Nz-1;z++)
    {
    for(y=1;y<Ny-1;y++)
      {
      for(x=1;x<Nx-1;x++)
        {
        ic=(Ny*z+y)*Nx+x;
        ie=ic+1;
        iw=ic-1;
        in=ic+Nx;
        is=ic-Nx;
        iu=ic+Ny*Nx;
        il=ic-Ny*Nx;
        Cc=C[ic];
        Ce=C[ie];
        Cn=C[in];
        Cs=C[is];
        Cw=C[iw];
        Cu=C[iu];
        Cl=C[il];
        Fc=F[ic];
        Fe=F[ie];
        Fn=F[in];
        Fs=F[is];
        Fw=F[iw];
        Fu=F[iu];
        Fl=F[il];
        Uc=U[ic];
        Vc=V[ic];
        Wc=W[ic];
        Ca=(Cw+Cc)/2.;
        Cb=(Ce+Cc)/2.;
        Cd=(Cn+Cc)/2.;
        Cf=(Cs+Cc)/2.;
        Cg=(Cu+Cc)/2.;
        Ch=(Cl+Cc)/2.;
        Da=(Dx[iw]+Dx[ic])/2.;
        Db=(Dx[ie]+Dx[ic])/2.;
        Dd=(Dy[in]+Dy[ic])/2.;
        Df=(Dy[is]+Dy[ic])/2.;
        Dg=(Dz[iu]+Dz[ic])/2.;
        Dh=(Dz[il]+Dz[ic])/2.;
        dFCdXa=(Fc*Cc-Fw*Cw)/dX;
        dFCdXb=(Fe*Ce-Fc*Cc)/dX;
        dFCdYd=(Fn*Cn-Fc*Cc)/dY;
        dFCdYf=(Fc*Cc-Fs*Cs)/dY;
        dFCdZg=(Fu*Cu-Fc*Cc)/dZ;
        dFCdZh=(Fc*Cc-Fl*Cl)/dZ;
        d2DFCdX2=(Db*dFCdXb-Da*dFCdXa)/dX;
        d2DFCdY2=(Dd*dFCdYd-Df*dFCdYf)/dY;
        d2DFCdZ2=(Dg*dFCdZg-Dh*dFCdZh)/dZ;
        if(Uc>0.)
          UdFCdX=Uc*(Fc*Cc-Fw*Cw)/dX;
        else
          UdFCdX=Uc*(Fe*Ce-Fc*Cc)/dX;
        if(Vc>0.)
          VdFCdY=Vc*(Fc*Cc-Fs*Cs)/dY;
        else
          VdFCdY=Vc*(Fn*Cn-Fc*Cc)/dY;
        if(Wc>0.)
          WdFCdZ=Wc*(Fc*Cc-Fl*Cl)/dZ;
        else
          WdFCdZ=Wc*(Fu*Cu-Fc*Cc)/dZ;
        dCdt[ic]=(d2DFCdX2+d2DFCdY2+d2DFCdZ2-UdFCdX-VdFCdY-WdFCdZ)/Fc;
        }
      }
    }
  for(z=1;z<Nz-1;z++)
    {
    for(y=1;y<Ny-1;y++)
      {
      for(x=1;x<Nx-1;x++)
        {
        ic=(Ny*z+y)*Nx+x;
        C[ic]+=dt*dCdt[ic];
        }
      }
    }
  for(y=0,x=1;x<Nx-1;x++)
    for(z=1;z<Nz-1;z++)
      C[(Ny*z+y)*Nx+x]=C[(Ny*z+y+1)*Nx+x];
  for(y=Ny-1,x=1;x<Nx-1;x++)
    for(z=1;z<Nz-1;z++)
      C[(Ny*z+y)*Nx+x]=C[(Ny*z+y-1)*Nx+x];
  for(x=0,y=1;y<Ny-1;y++)
    for(z=1;z<Nz-1;z++)
      C[(Ny*z+y)*Nx+x]=C[(Ny*z+y)*Nx+x+1];
  for(x=Nx-1,y=1;y<Ny-1;y++)
    for(z=1;z<Nz-1;z++)
      C[(Ny*z+y)*Nx+x]=C[(Ny*z+y)*Nx+x-1];
  for(z=0,x=1;x<Nx-1;x++)
    for(y=1;y<Ny-1;y++)
      C[(Ny*z+y)*Nx+x]=C[(Ny*(z+1)+y)*Nx+x];
  for(z=Nz-1,x=1;x<Nx-1;x++)
    for(y=1;y<Ny-1;y++)
      C[(Ny*z+y)*Nx+x]=C[(Ny*(z-1)+y)*Nx+x];
  x=y=z=0;
  C[node(x,y,z)]=(C[node(x+1,y,z)]+C[node(x,y+1,z)]+C[node(x,y,z+1)])/3.;
  x=y=0;z=Nz-1;
  C[node(x,y,z)]=(C[node(x+1,y,z)]+C[node(x,y+1,z)]+C[node(x,y,z-1)])/3.;
  x=z=0;y=Ny-1;
  C[node(x,y,z)]=(C[node(x+1,y,z)]+C[node(x,y-1,z)]+C[node(x,y,z+1)])/3.;
  x=0;y=Ny-1;z=Nz-1;
  C[node(x,y,z)]=(C[node(x+1,y,z)]+C[node(x,y-1,z)]+C[node(x,y,z-1)])/3.;
  x=Nx-1;y=z=0;
  C[node(x,y,z)]=(C[node(x-1,y,z)]+C[node(x,y+1,z)]+C[node(x,y,z+1)])/3.;
  x=Nx-1;y=0;z=Nz-1;
  C[node(x,y,z)]=(C[node(x-1,y,z)]+C[node(x,y+1,z)]+C[node(x,y,z-1)])/3.;
  x=Nx-1;y=Ny-1;z=0;
  C[node(x,y,z)]=(C[node(x-1,y,z)]+C[node(x,y-1,z)]+C[node(x,y,z+1)])/3.;
  x=Nx-1;y=Ny-1;z=Nz-1;
  C[node(x,y,z)]=(C[node(x-1,y,z)]+C[node(x,y-1,z)]+C[node(x,y,z-1)])/3.;
  }

void FileResults(char*name)
  {
  int i,x,y,z;
  double logCm,X,Y,Z;
  FILE*fp;
  printf("%s\n",name);
  if((fp=fopen(name,"wt"))==NULL)
    Abort(__LINE__,"can't create output file");
  fprintf(fp,"%i\n",Nx);
  for(x=0;x<Nx;x++)
    {
    X=(x-Nx/2)*dX;
    fprintf(fp,"%lG\n",X);
    }
  fprintf(fp,"%i\n",Ny);
  for(y=0;y<Ny;y++)
    {
    Y=(y-Ny/2)*dY;
    fprintf(fp,"%lG\n",Y);
    }
  fprintf(fp,"%i\n",Nz);
  for(z=0;z<Nz;z++)
    {
    Z=(z-Nz/2)*dZ;
    fprintf(fp,"%lG\n",Z);
    }
  fprintf(fp,"%i\n",Nx*Ny*Nz);
  logCm=log(Cm);
  for(i=z=0;z<Nz;z++)
    for(y=0;y<Ny;y++)
      for(x=0;x<Nx;x++,i++)
        if(C[i]>Cm)
          fprintf(fp,"%lG\n",log(C[i]));
        else
          fprintf(fp,"%lG\n",logCm);
  fclose(fp);
  }

int main(int argc,char**argv,char**envp)
  {
  char name[]="PCE3D000.TB3";
  int i,j,n;
  double atx,aty,atz,dx,dy,dz,dt,dtx,dty,dtz,t,ux,vx,wx;
  printf("3D Transient Diffusion + Dispersion + Advection\n");
  InitializePorosity();
  InitializeDiffusion();
  InitializeVelocity();
  dx=Maximum(Dx,Nx*Ny*Nz);
  dy=Maximum(Dy,Nx*Ny*Nz);
  dz=Maximum(Dz,Nx*Ny*Nz);
  printf("Dx%lG, Dy%lG, Dz%lG cm/sec\n",dx*cm_per_m*cm_per_m/sec_per_yr,dy*cm_per_m*cm_per_m/sec_per_yr,dz*cm_per_m*cm_per_m/sec_per_yr);
  dtx=dX*dX/dx/2.;
  dty=dY*dY/dy/2.;
  dtz=dZ*dZ/dz/2.;
  printf("dX/Dx/2=%lG, dY/Dy/2=%lG, dZ/Dz/2=%lG yr\n",dtx,dty,dtz);
  ux=fmax(fabs(Minimum(U,Nx*Ny*Nz)),Maximum(U,Nx*Ny*Nz));
  vx=fmax(fabs(Minimum(V,Nx*Ny*Nz)),Maximum(V,Nx*Ny*Nz));
  wx=fmax(fabs(Minimum(W,Nx*Ny*Nz)),Maximum(W,Nx*Ny*Nz));
  printf("U%lG, V%lG, W%lG m/day\n",ux/day_per_yr,vx/day_per_yr,wx/day_per_yr);
  atx=dX/ux/6.;
  aty=dY/vx/6.;
  atz=dZ/wx/6.;
  printf("dX/U/6=%lG, dY/V/6=%lG, dZ/W/6=%lG yr\n",atx,aty,atz);
  dt=fmin(dtx,fmin(dty,fmin(dtz,fmin(atx,fmin(aty,atz)))));
  if(dt>1.)
    dt=1.;
  else if(dt<1.)
    dt=1./(int)(1./dt+1.5);
  printf("t=%lG yr\n",dt);
  InitializeConcentration();
  Co=Maximum(C,Nx*Ny*Nz);
  printf("0 years C%.0lf ",Co);
  FileResults(name);
  n=(int)(5./dt+0.5);
  for(t=j=i=0;t<1800.;i++)
    {
    AdvanceSolution(dt);
    Cx=Maximum(C,Nx*Ny*Nz);
    if(Cx>Co)
      Abort(__LINE__,"time step is too large\nsolution is unstable");
    t+=dt;
    if((i+1)%n==0)
      {
      printf("%.0lf years C%.0lf ",t,Cx);
      sprintf(name,"PCE3D%03d.TB3",++j);
      FileResults(name);
      if(Cx<Co/10.)
        break;
      }
    }
  return(0);
  }
